/**
 * $Id$ 
 * 
 * Copyright (c) 2008, Jens Elkner.
 * All rights reserved.
 *
 * Code licensed under the BSD License: http://developer.yahoo.net/yui/license.txt
 */
package com.yahoo.platform.yui.compressor;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;

import org.apache.tools.ant.types.Parameter;
import org.mozilla.javascript.ErrorReporter;
import org.mozilla.javascript.EvaluatorException;

/**
 * A JavaScript reader, which compresses the input per default.
 * NOTE: The reader always reads in the complete stream as soon as a 
 * {@link #read(char[], int, int)} or {@link #skip(long)} or {@link #mark(int)}
 * is invoked on this filter and than buffers the result for consecutive 
 * operations. So you may need to invoke ant with something like {@code -Xmx256m}
 * for large files.
 * <p>
 * Dueto the bad design of {@link org.apache.tools.ant.filters.BaseFilterReader} 
 * this filter is not able to use the ant logging system and thus emits all
 * error messages to stderr, warnings and verbose output to stdout.
 * 
 * @author 	Jens Elkner
 * @version	$Revision$
 */
public class JsFilter
	extends YUIFilter
{
	private boolean verbose;
	private boolean munge;
	private boolean preserveSemi;
	private boolean optimize;

	/**
	 * Create a new Javascript compressing filter, which breaks the resulting 
	 * String into lines of max. {@var linebreak} characters.
	 * @param in			the reader, which provides the CSS to compress.
	 * @param linebreak		max. length of resulting lines. A negative value 
	 * 			implies unlimited length.
	 * @param verbose If {@code true}, display informational messages and warnings.
	 * @param obfuscate If {@code true} (default), obfuscate the code as well.
	 * @param preserveSemi If {@code true}, preserve all semicolons.
	 * @param optimze If {@code true} (default), code optimization gets enabled.
	 */
	public JsFilter(Reader in, int linebreak, boolean verbose, boolean obfuscate,
		boolean preserveSemi, boolean optimze) 
	{
		super(in, linebreak);
		this.verbose = verbose;
		this.munge = obfuscate;
		this.preserveSemi = preserveSemi;
		this.optimize = optimze;
	}
	
	/**
	 * Create a Javascript compressing filter, which breaks the resulting 
	 * String into lines of max. {@var linebreak} characters. Optimization and 
	 * obfuscation is enabled, verbose messages and preserving all semicolons is
	 * disabled.
	 * @param in			the reader, which provides the CSS to compress.
	 * @param linebreak		max. length of resulting lines. A negative value 
	 * 			implies unlimited length.
	 */
	public JsFilter(Reader in, int linebreak) {
		this(in, linebreak, false, true, false, true);
	}
	
	/**
	 * Create a Javascript compressing filter.
	 * Optimization and obfuscation is enabled, verbose messages and preserving 
	 * all semicolons is disabled.
	 * @param in	where to read the uncompressed javascript code.
	 */
	public JsFilter(Reader in) {
		this(in, -1);
	}

	/**
	 * Creates an YUI javascript compressor, parses the input stream and 
	 * provides the results in the returned reader.
	 */
	@Override
	protected StringReader getResults() throws IOException {
		JavaScriptCompressor compressor = new JavaScriptCompressor(in,
			new ErrorReporter() {
				public void warning(String message, String sourceName,
					int line, String lineSource, int lineOffset)
				{
					if (line < 0) {
						System.out.println(message);
					} else {
						System.out.println(line + ':' + lineOffset + ':' + message);
					}
				}

				public void error(String message, String sourceName,
					int line, String lineSource, int lineOffset)
				{
					if (line < 0) {
						System.err.println(message);
					} else {
						System.err.println(line + ':' + lineOffset + ':' + message);
					}
				}

				public EvaluatorException runtimeError(String message,
					String sourceName, int line, String lineSource,
					int lineOffset)
				{
					error(message, sourceName, line, lineSource, lineOffset);
					return new EvaluatorException(message);
				}
			}
		);
		StringWriter out = new StringWriter(4096);
		compressor.compress(out, linebreak, munge, verbose, preserveSemi, !optimize);
		return new StringReader(out.toString());
	}

	/**
	 * Additional parameters for this filter.
	 * <ul>
	 * 	<li>verbose .. If {@code true}, display informational messages and warnings</li>
	 * 	<li>munge .. If {@code true} (default), obfuscate the code as well.</li>
	 * 	<li>preserveSemi .. If {@code true}, preserve all semicolons.</li>
	 * 	<li>optimize .. If {@code true} (default), code optimization gets enabled.</li>
	 * </ul>
	 * Parameters must be set, before any operation on the stream occurs.
	 * @see YUIFilter#setParameters(Parameter[])
	 */
	@Override
	public void setParameters(Parameter[] parameters) {
		if (parameters == null) {
			return;
		}
		ArrayList<Parameter> params = setParams(parameters);
		if (initialized) {
			return;
		}
		for (Parameter param : params) {
			String tmp = param.getName();
			if (tmp.equals("verbose")) {
				tmp = param.getValue();
				verbose = 
					tmp.equalsIgnoreCase("true") || tmp.equalsIgnoreCase("on");
			} else if (tmp.equals("munge")) {
				tmp = param.getValue();
				munge = 
					tmp.equalsIgnoreCase("true") || tmp.equalsIgnoreCase("on");
			} else if (tmp.equals("preserveSemi")) {
				tmp = param.getValue();
				preserveSemi = 
					tmp.equalsIgnoreCase("true") || tmp.equalsIgnoreCase("on");
			} else if (tmp.equals("optimize")) {
				tmp = param.getValue();
				optimize = 
					tmp.equalsIgnoreCase("true") || tmp.equalsIgnoreCase("on");
			} else {
				System.err.println("Unknown parameter '" + tmp + "' ignored");
			}
		}
	}
	
}
