/**
 * $Id$ 
 * 
 * Copyright (c) 2008, Jens Elkner.
 * All rights reserved.
 *
 * Code licensed under the BSD License: http://developer.yahoo.net/yui/license.txt
 */
package com.yahoo.platform.yui.compressor;

import java.io.FilterReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayList;

import org.apache.tools.ant.types.Parameter;
import org.apache.tools.ant.types.Parameterizable;

/**
 * The base of a YUI compressor filter.
 * 
 * @author 	Jens Elkner
 * @version	$Revision$
 */
public abstract class YUIFilter
	extends FilterReader
	implements Parameterizable
{
	/** max line length of the result */
	protected int linebreak = 0;
	/** if {code true}, compression is enabled. */
	protected boolean enabled = true;
	/** indicates, whether operations on the stream were already made, i.e.
	 * changing the {@link #enabled} parameter should not be allowed anymore */
	protected boolean initialized = false;
	/** the internal reader used to provide the results */
	private StringReader sin;

	/**
	 * Create a new YUI compressing filter, which breaks the resulting String 
	 * into lines of max. {@var linebreak} characters.
	 * @param in			the reader, which provides the CSS to compress.
	 * @param linebreak		max. length of resulting lines. A negative value 
	 * 			implies unlimited length.
	 */
	public YUIFilter(Reader in, int linebreak) {
		super(in);
		this.linebreak = linebreak;
	}

	/**
	 * If {{@link #enabled} is set to {@code true}, create the YUI compressor 
	 * (Javascript or CSS), parse the input and returns the reader, which 
	 * contains the results of compression. The reader will be closed 
	 * automatically, when this filter gets closed.
	 * 
	 * @return the reader, which contains the results the compressor produced.
	 * 
	 * @throws IOException
	 */
	protected abstract StringReader getResults() throws IOException;

	/**
	 * Parameters for this filter.
	 * <ul>
	 * 	<li>linebreak .. max. length of resulting lines. . A negative value 
	 * 		implies unlimited length.</li>
	 * 	<li>enable .. If {@code true} (default), compression is enabled.</li>
	 * 	<li>debug .. If {@code true}, do nothing - i.e. passthrough as is. 
	 * 		Actually it is the opposite of {@code enable} to make ant scripting 
	 * 		easier.</li>
	 * </ul>
	 * Parameters must be set, before any operation on the stream occurs.
	 * @param parameters parameters to set
	 * @return a list of all unprocessed/unknown parameters
	 */
	protected ArrayList<Parameter> setParams(Parameter[] parameters) {
		if (parameters == null) {
			return null;
		}
		if (initialized) {
			System.err.println("Too late to modify the resulting stream.");
			return null;
		}
		ArrayList<Parameter> params = new ArrayList<Parameter>(parameters.length);
		for (int i=0; i < parameters.length; i++) {
			String tmp = parameters[i].getName();
			if (tmp.equals("linebreak")) {
				try {
					int x = Integer.parseInt(parameters[i].getValue(), 10);
					linebreak = x;
				} catch (Exception e) {
					System.err.println("Invalid value '" 
						+ parameters[i].getValue() + "' for linebreak ignored");
				}
			} else if (tmp.equals("enable")) {
				tmp = parameters[i].getValue();
				enabled = tmp.equalsIgnoreCase("true") || tmp.equalsIgnoreCase("on");
			} else if (tmp.equals("debug")) {
				tmp = parameters[i].getValue();
				enabled = ! (tmp.equalsIgnoreCase("true") || tmp.equalsIgnoreCase("on"));
			} else {
				params.add(parameters[i]);
			}
		}
		return params;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int read() throws IOException {
		initialized = true;
		if (!enabled) {
			return in.read();
		}
		if (sin == null) {
			sin = getResults();
		}
		return sin.read();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int read(char cbuf[], int off, int len)
		throws IOException
	{
		initialized = true;
		if (!enabled) {
			return in.read(cbuf, off, len);
		}
		if (sin == null) {
			sin = getResults();
		}
		return sin.read(cbuf, off, len);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public long skip(long n) throws IOException {
		initialized = true;
		if (!enabled) {
			return in.skip(n);
		}
		if (sin == null) {
			sin = getResults();
		}
		return sin.skip(n);
	}

	/**
	 * {@inheritDoc}
	 * @throws IOException 
	 */
	@Override
	public boolean ready() throws IOException {
		initialized = true;
		return enabled ? true : in.ready();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean markSupported() {
		initialized = true;
		return enabled ? true : in.markSupported();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void mark(int readAheadLimit) throws IOException {
		initialized = true;
		if (!enabled) {
			in.mark(readAheadLimit);
			return;
		}
		if (sin == null) {
			sin = getResults();
		}
		sin.mark(readAheadLimit);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void reset() throws IOException {
		initialized = true;
		if (!enabled) {
			in.reset();
			return;
		}
		if (sin != null) {
			sin.reset();
		}
	}

	/**
	 * Close this filter as well as all underlying and internal readers.
	 */
	@Override
	public void close() throws IOException {
		if (sin != null) {
			sin.close();
		}
		super.close();
	}
}